from enum import Enum
from pathlib import Path
from typing import Any, Dict

import tomli


def pwd() -> Path:
    return Path(".")


def load_pyproject() -> Dict[str, Any]:
    """Load Configuration from pyproject.toml file."""
    pyproject: Path = Path("pyproject.toml")
    return tomli.loads(pyproject.read_text()) if pyproject.exists() else {}


def ls(path: str):
    yield from pwd().glob(path)


def readline(path: str):
    file = pwd() / Path(path)
    return file.read_text(encoding="utf-8")


class Bcolors(str, Enum):
    """A Enum for colors using ANSI escape sequences.
    Reference:
    - https://stackoverflow.com/questions/287871

    """

    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKCYAN = "\033[96m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    OK = "\033[92m"
    INFO = "\033[94m"
    ERROR = "\033[91m"


class Level(str, Enum):
    """An Enum for notification levels."""

    OK = "OK"
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"


def make_color(message: str, level: Level) -> str:
    """Print the message with a color for the corresponding level."""
    return f"{Bcolors[level]}{Bcolors.BOLD}{level}: {message}{Bcolors.ENDC}"



from __future__ import annotations

import os
import re
import subprocess
import sys
from collections import defaultdict
from pathlib import Path
from typing import Any, Dict, List, Optional

import click

from .git import CommitLog

BUMP_VERSION = (("bump", ":bookmark:"),)  # 🔖 :bookmark:

BUMP_REGEX = (
    r"(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+)"
    r"(\.(?P<prekind>a|alpha|b|beta|d|dev|rc)(?P<pre>\d+))?"
    r"(\.(?P<postkind>post)(?P<post>\d+))?"
)

BUMP_VERSION_CONFIG = r"""[bumpversion]
current_version = {version}
commit = True
tag = False
parse = ^
    (?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+)
    (\.(?P<prekind>a|alpha|b|beta|d|dev|rc)(?P<pre>\d+))?
    (\.(?P<postkind>post)(?P<post>\d+))?
serialize =
    {{major}}.{{minor}}.{{patch}}.{{prekind}}{{pre}}.{{postkind}}{{post}}
    {{major}}.{{minor}}.{{patch}}.{{prekind}}{{pre}}
    {{major}}.{{minor}}.{{patch}}.{{postkind}}{{post}}
    {{major}}.{{minor}}.{{patch}}
message = :bookmark: Bump up to version {{current_version}} -> {{new_version}}.

[bumpversion:part:prekind]
optional_value = _
values =
    _
    dev
    a
    b
    rc

[bumpversion:part:postkind]
optional_value = _
values =
    _
    post

[bumpversion:file:{file}]

[bumpversion:file:{changelog}]
search = ## Latest Changes
replace = ## {{new_version}}
"""

cli_vs: click.Command


def generate_group_commit_log() -> Dict[str, List[CommitLog]]:
    """Generate Group of the Commit Logs"""
    from .git import get_commit_logs

    group_logs: Dict[str, List[CommitLog]] = defaultdict(list)
    for log in get_commit_logs():
        group_logs[log.msg.mtype].append(log)
    return {
        k: sorted(v, key=lambda x: x.date, reverse=True)
        for k, v in group_logs.items()
    }


# TODO: add new style of changelog file
# TODO: add parameter that able to write after release version like
#  hot-changes commit
def writer_changelog(file: str):
    """Write Commit logs to the changelog file."""
    group_logs: Dict[str, List[CommitLog]] = generate_group_commit_log()

    with Path(file).open(encoding="utf-8") as f_changes:
        changes = f_changes.read().splitlines()

    writer = Path(file).open(mode="w", encoding="utf-8", newline="")
    skip_line: bool = True
    written: bool = False
    for line in changes:
        if line.startswith("## Latest Changes"):
            skip_line = False

        if re.match(rf"## {BUMP_REGEX}", line):
            if not written:
                writer.write(f"## Latest Changes{os.linesep}{os.linesep}")
                written = True
            skip_line = True

        if skip_line:
            writer.write(line + os.linesep)
        elif written:
            continue
        else:
            from .git import COMMIT_PREFIX_TYPE

            linesep = os.linesep
            if any(cpt[0] in group_logs for cpt in COMMIT_PREFIX_TYPE):
                linesep = f"{os.linesep}{os.linesep}"

            writer.write(f"## Latest Changes{linesep}")

            for cpt in COMMIT_PREFIX_TYPE:
                if cpt[0] in group_logs:
                    writer.write(f"### {cpt[0]}{os.linesep}{os.linesep}")
                    for log in group_logs[cpt[0]]:
                        writer.write(
                            f"- {log.msg.content} (_{log.date:%Y-%m-%d}_)"
                            f"{os.linesep}"
                        )
                    writer.write(os.linesep)
            written = True
    writer.close()


def bump2version(
    action: str,
    file: str,
    changelog_file: str,
    ignore_changelog: bool = False,
    dry_run: bool = False,
):
    from .git import merge2latest_commit

    with Path(".bumpversion.cfg").open(mode="w", encoding="utf-8") as f_bump:
        f_bump.write(
            BUMP_VERSION_CONFIG.format(
                file=file,
                version=current_version(file),
                changelog=changelog_file,
            )
        )
    if not ignore_changelog:
        writer_changelog(file=changelog_file)
    subprocess.run(["git", "add", "-A"])
    subprocess.run(
        [
            "git",
            "commit",
            "-m",
            "build: add bump2version config file",
            "--no-verify",
        ]
    )
    subprocess.run(
        [
            "bump2version",
            action,
            "--commit-args=--no-verify",
        ]
        + (["--list", "--dry-run"] if dry_run else [])
    )
    writer_changelog(file=changelog_file)
    Path(".bumpversion.cfg").unlink(missing_ok=False)
    merge2latest_commit(no_verify=True)
    subprocess.run(
        [
            "git",
            "reset",
            "--soft",
            "HEAD~1",
        ],
        stderr=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
    )
    with Path(".git/COMMIT_EDITMSG").open(encoding="utf-8") as f_msg:
        raw_msg = f_msg.read().splitlines()
    subprocess.run(
        [
            "git",
            "commit",
            "--amend",
            "-m",
            raw_msg[0],
            "--no-verify",
        ],
        stderr=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
    )
    return 0


def current_version(file: str) -> str:
    with Path(file).open(encoding="utf-8") as f:
        if search := re.search(BUMP_REGEX, f.read()):
            return search[0]
    raise NotImplementedError(f"{file} does not implement version value.")


def load_project() -> Dict[str, Any]:
    from .utils import load_pyproject

    return load_pyproject().get("project", {})


def load_config() -> Dict[str, Any]:
    from .utils import load_pyproject

    return load_pyproject().get("tool", {}).get("utils", {}).get("version", {})


@click.group(name="vs")
def cli_vs():
    """Version commands"""
    pass


@cli_vs.command()
def conf():
    """Return Configuration for Bump version"""
    sys.exit(load_config())


@cli_vs.command()
@click.option("-f", "--file", type=click.Path(exists=True))
def changelog(file: Optional[str]):
    """Make Changelogs file"""
    if not file:
        file = load_config().get("changelog", None) or "CHANGELOG.md"
    writer_changelog(file)
    sys.exit(0)


@cli_vs.command()
@click.option("-f", "--file", type=click.Path(exists=True))
def current(file: str):
    """Return Current Version"""
    if not file:
        file = load_config().get("version", None) or (
            f"./{load_project().get('name', 'unknown')}/__about__.py"
        )
    sys.exit(current_version(file))


@cli_vs.command()
@click.argument("action", type=click.STRING)
@click.option("-f", "--file", type=click.Path(exists=True))
@click.option("-c", "--changelog-file", type=click.Path(exists=True))
@click.option("--ignore-changelog", is_flag=True)
@click.option("--dry-run", is_flag=True)
def bump(
    action: str,
    file: Optional[str],
    changelog_file: Optional[str],
    ignore_changelog: bool,
    dry_run: bool,
):
    """Bump Version"""
    if not file:
        file = load_config().get("version", None) or (
            f"./{load_project().get('name', 'unknown')}/__about__.py"
        )
    if not changelog_file:
        changelog_file = load_config().get("changelog", None) or "CHANGELOG.md"
    sys.exit(
        bump2version(action, file, changelog_file, ignore_changelog, dry_run)
    )


if __name__ == "__main__":
    cli_vs.main()



import unittest

from click.testing import CliRunner

from clishelf.cli import say


class MainTestCase(unittest.TestCase):
    def test_hello_world(self):
        runner = CliRunner()
        result = runner.invoke(say)
        self.assertEqual(result.exit_code, 1)
        self.assertEqual(result.output, "Hello World\n")



from __future__ import annotations

import subprocess
import sys
from pathlib import Path

import click

from .git import cli_git
from .version import cli_vs


@click.group()
def cli():
    """A simple command line tool."""
    pass  # pragma: no cover.


@cli.command()
@click.option(
    "-p",
    "--path",
    type=click.Path(exists=True, resolve_path=True),
    default=".",
)
def ls(path: str):
    """List files in Current Path"""
    for file in Path(path).glob("*"):
        print(f"> {file.resolve()}")
    sys.exit(0)


@cli.command()
def say():
    """Say Hello World"""
    sys.exit("Hello World")


@cli.command()
@click.option(
    "-m",
    "--module",
    type=click.STRING,
    default="pytest",
)
@click.option(
    "-h",
    "--html",
    is_flag=True,
)
def cove(module: str, html: bool):
    """Run coverage"""
    subprocess.run(["coverage", "run", "--m", module, "tests"])
    subprocess.run(["coverage", "combine"])
    subprocess.run(["coverage", "report", "--show-missing"])
    if html:
        subprocess.run(["coverage", "html"])


def main() -> None:
    cli.add_command(cli_git)
    cli.add_command(cli_vs)
    cli.main()


if __name__ == "__main__":
    main()



from __future__ import annotations

import os
import re
import subprocess
import sys
from dataclasses import InitVar, dataclass
from datetime import date, datetime
from pathlib import Path
from typing import List, Optional, Tuple

import click

from .utils import (
    Level,
    make_color,
)

cli_git: click.Command

PROJECT_IDS = ["SLO"]
BRANCH_TYPES = ["feature", "bug", "hot"]

REGEX_PROJECT_IDS = "|".join(PROJECT_IDS)
REGEX_BRANCH_TYPES = "|".join(BRANCH_TYPES)

# Should contain a capturing group to extract the reference:
REGEX_BRANCH = rf"^(?:{REGEX_BRANCH_TYPES})/((?:{REGEX_PROJECT_IDS})-[\d]{{1,5}})-[a-z]+(?:-[a-z]+)*$"

# Should contain a capturing group to extract the reference (note the dot at the end
# is optional as this script will add it automatically for us):
REGEX_MESSAGE = rf"^((?:{REGEX_PROJECT_IDS})-[\d]{{1,5}}): .+\.?$"

# No capturing group. Just checking for the bare minimum:
REGEX_BASIC_MESSAGE = "^.+$"

REGEX_COMMIT_MESSAGE = r"(?P<prefix>\w+)(?:\((?P<topic>\w+)\))?: (?P<header>.+)"

# These branch names are not validated with this same rules (permissions should be configured
# on the server if you want to prevent pushing to any of these):
BRANCH_EXCEPTIONS = [
    "feature",
    "dev",
    "main",
    "stable",
    "hotfix",  # for quickly fixing critical issues, usually with a temporary solution
    "bugfix",  # for fixing a bug
    "feature",  # for adding, removing or modifying a feature
    "test",  # for experimenting something which is not an issue
    "wip",  # for a work in progress
]

COMMIT_PREFIX = (
    ("feat", "Features", ":dart:"),  # 🎯, 📋 :clipboard:
    ("hotfix", "Fix Bugs", ":fire:"),  # 🔥
    ("fixed", "Fix Bugs", ":gear:"),  # ⚙️, 🛠️ :hammer_and_wrench:
    ("fix", "Fix Bugs", ":gear:"),  # ⚙️, 🛠️ :hammer_and_wrench:
    ("docs", "Documents", ":page_facing_up:"),  # 📄, 📑 :bookmark_tabs:
    ("style", "Code Changes", ":art:"),  # 🎨, 📝 :memo:, ✒️ :black_nib:
    ("refactored", "Code Changes", ":construction:"),  # 🚧, 💬 :speech_balloon:
    ("refactor", "Code Changes", ":construction:"),  # 🚧, 💬 :speech_balloon:
    ("perf", "Code Changes", ":chart_with_upwards_trend:"),  # 📈, ⌛ :hourglass:
    ("test", "Code Changes", ":test_tube:"),  # 🧪, ⚗️ :alembic:
    ("build", "Build & Workflow", ":toolbox:"),  # 🧰, 📦 :package:
    ("workflow", "Build & Workflow", ":rocket:"),  # 🚀, 🕹️ :joystick:
)

COMMIT_PREFIX_TYPE = (
    ("Features", ":clipboard:"),  # 📋
    ("Code Changes", ":black_nib:"),  # ✒️
    ("Documents", ":bookmark_tabs:"),  # 📑
    ("Fix Bugs", ":hammer_and_wrench:"),  # 🛠️
    ("Build & Workflow", ":package:"),  # 📦
)


@dataclass
class CommitMsg:
    content: InitVar[str]
    mtype: InitVar[str] = None
    body: str = None  # Mark new-line with |

    def __str__(self):
        return f"{self.mtype}: {self.content}"

    def __post_init__(self, content: str, mtype: str):
        self.content: str = self.__prepare_msg(content)
        if not mtype:
            self.mtype: str = self.__gen_msg_type()

    def __gen_msg_type(self) -> str:
        if s := re.search(r"^:\w+:\s(?P<prefix>\w+):", self.content):
            prefix: str = s.groupdict()["prefix"]
            return next(
                (cp[1] for cp in COMMIT_PREFIX if prefix == cp[0]),
                "Code Changes",
            )
        return "Code Changes"

    @property
    def mtype_icon(self):
        return next(
            (cpt[1] for cpt in COMMIT_PREFIX_TYPE if cpt[0] == self.mtype),
            ":black_nib:",
        )

    @staticmethod
    def __prepare_msg(content: str) -> str:
        if re.match(r"^:\w+:", content):
            return content

        prefix, content = (
            content.split(":", maxsplit=1)
            if ":" in content
            else ("refactor", content)
        )
        icon: str = ""
        for cp in COMMIT_PREFIX:
            if prefix == cp[0]:
                icon = f"{cp[2]} "
        return f"{icon}{prefix}: {content.strip()}"


@dataclass(frozen=True)
class CommitLog:
    hash: str
    date: date
    msg: CommitMsg
    author: str

    def __str__(self) -> str:
        return "|".join(
            (
                self.hash,
                self.date.strftime("%Y-%m-%d"),
                self.msg.content,
                self.author,
            )
        )


def validate_for_warning(
    lines: List[str],
) -> List[str]:
    subject: str = lines[0]
    results: List[str] = []

    # RULE 02: Limit the subject line to 50 characters
    if len(subject) <= 20 or len(subject) > 50:
        results.append(
            "There should be between 21 and 50 characters in the commit title."
        )
    if len(lines) <= 2:
        results.append("There should at least 3 lines in your commit message.")

    # RULE 01: Separate subject from body with a blank line
    if lines[1].strip() != "":
        results.append(
            "There should be an empty line between "
            "the commit title and body."
        )

    if not lines[0].strip().endswith("."):
        lines[0] = f"{lines[0].strip()}."
        results.append("There should not has dot in the end of commit message.")
    return results


def validate_commit_msg(
    lines: List[str],
) -> Tuple[List[str], Level]:
    if not lines:
        return (
            ["Please supply commit message without start with ``#``."],
            Level.ERROR,
        )

    rs = validate_for_warning(lines)
    if rs:
        return rs, Level.WARNING

    has_story_id: bool = False
    for line in lines[1:]:
        # RULE 06: Wrap the body at 72 characters
        if len(line) > 72:
            rs.append("The commit body should wrap at 72 characters.")

        if line.startswith("["):
            has_story_id = True

    if not has_story_id:
        rs.append("Please add a Story ID in the commit message.")

    if not rs:
        return (
            ["The commit message has the required pattern."],
            Level.OK,
        )
    return rs, Level.WARNING


def get_branch_name() -> str:
    return (
        subprocess.check_output(["git", "rev-parse", "--abbrev-ref", "HEAD"])
        .decode(sys.stdout.encoding)
        .strip()
    )


def get_latest_tag(default: bool = True) -> Optional[str]:
    try:
        return (
            subprocess.check_output(
                ["git", "describe", "--tags", "--abbrev=0"],
                stderr=subprocess.DEVNULL,
            )
            .decode(sys.stdout.encoding)
            .strip()
        )
    except subprocess.CalledProcessError:
        if default:
            from ..__about__ import __version__

            return f"v{__version__}"
        return None


def prepare_commit_logs(tag2head: str):
    """Prepare contents logs to List of commit log."""
    results: List = []
    prepare: List[str] = []
    for line in (
        subprocess.check_output(
            [
                "git",
                "log",
                tag2head,
                "--pretty=format:%h|%ad|%an%n%s%n%b%-C()%n(END)",
                "--date=short",
            ]
        )
        .decode(sys.stdout.encoding)
        .strip()
        .splitlines()
    ):
        if line == "(END)":
            results.append(prepare)
            prepare = []
            continue
        prepare.append(line)
    return results


def get_commit_logs(
    tag: Optional[str] = None,
    all_logs: bool = False,
) -> List[CommitLog]:
    if tag:
        tag2head: str = f"{tag}..HEAD"
    elif all_logs or not (tag := get_latest_tag(default=False)):
        tag2head = "HEAD"
    else:
        tag2head = f"{tag}..HEAD"
    msgs: List[CommitLog] = []
    for _ in prepare_commit_logs(tag2head):
        if "Merge" in _[1]:
            continue

        _s: List[str] = _[0].split("|")
        msgs.append(
            CommitLog(
                hash=_s[0],
                date=datetime.strptime(_s[1], "%Y-%m-%d"),
                msg=CommitMsg(content=_[1], body="|".join(_[2:])),
                author=_s[2],
            )
        )
    return msgs


def merge2latest_commit(no_verify: bool = False):
    subprocess.run(
        ["git", "commit", "--amend", "--no-edit", "-a"]
        + (["--no-verify"] if no_verify else [])
    )


def get_latest_commit(
    file: Optional[str] = None,
    edit: bool = False,
    output_file: bool = False,
) -> List[str]:
    if file:
        with Path(file).open(encoding="utf-8") as f_msg:
            raw_msg = f_msg.read().splitlines()
    else:
        raw_msg = (
            subprocess.check_output(
                ["git", "log", "HEAD^..HEAD", "--pretty=format:%B"]
            )
            .decode(sys.stdout.encoding)
            .strip()
            .splitlines()
        )
    lines: List[str] = [
        msg for msg in raw_msg if not msg.strip().startswith("#")
    ]
    if lines[-1] != "":
        lines += [""]  # Add end-of-file line

    rss, level = validate_commit_msg(lines)
    for rs in rss:
        print(make_color(rs, level))
    if level not in (Level.OK, Level.WARNING):
        sys.exit(1)

    if edit:
        lines[0] = CommitMsg(content=lines[0]).content

    if file and output_file:
        with Path(file).open(mode="w", encoding="utf-8", newline="") as f_msg:
            f_msg.write(f"{os.linesep}".join(lines))
    return lines


def get_branch_ref(branch):
    match = re.findall(REGEX_BRANCH, branch)
    return match[0] if match and match[0] else None


@click.group(name="git")
def cli_git():
    """Extended Git commands"""
    pass


@cli_git.command()
def bn():
    """Show the Current Branch"""
    sys.exit(get_branch_name())


@cli_git.command()
def tl():
    """Show the Latest Tag"""
    sys.exit(get_latest_tag())


@cli_git.command()
@click.option("-t", "--tag", type=click.STRING, default=None)
@click.option("-a", "--all-logs", is_flag=True)
def cl(tag: Optional[str], all_logs: bool):
    """Show the Commit Logs from the latest Tag to HEAD"""
    sys.exit(
        "\n".join(str(x) for x in get_commit_logs(tag=tag, all_logs=all_logs)),
    )


@cli_git.command()
@click.option("-f", "--file", type=click.STRING, default=None)
@click.option("-l", "--latest", is_flag=True)
@click.option("-e", "--edit", is_flag=True)
@click.option("-o", "--output-file", is_flag=True)
@click.option("-p", "--prepare", is_flag=True)
def cm(
    file: Optional[str],
    latest: bool,
    edit: bool,
    output_file: bool,
    prepare: bool,
):
    """Show the latest Commit message"""
    if latest and not file:
        file = ".git/COMMIT_EDITMSG"
    if not prepare:
        print(
            "\n".join(get_latest_commit(file, edit, output_file)),
        )
        sys.exit(0)
    else:
        edit: bool = True
        cm_msg: str = "\n".join(get_latest_commit(file, edit, output_file))
        subprocess.run(
            [
                "git",
                "commit",
                "--amend",
                "-a",
                "--no-verify",
                "-m",
                cm_msg,
            ]
        )


@cli_git.command()
@click.option("--no-verify", is_flag=True)
def commit_previous(no_verify: bool):
    """Commit changes to the Previous Commit with same message"""
    merge2latest_commit(no_verify=no_verify)


@cli_git.command()
@click.option("-f", "--force", is_flag=True)
def commit_revert(force: bool):
    """Revert the latest Commit on this Local"""
    subprocess.run(["git", "reset", "HEAD^"])
    if force:
        subprocess.run(["git", "restore", "."])


@cli_git.command()
def clear_branch():
    """Clear Local Branches that sync from the Remote"""
    subprocess.run(
        ["git", "checkout", "main"],
        stderr=subprocess.DEVNULL,
        stdout=subprocess.DEVNULL,
    )
    subprocess.run(
        ["git", "remote", "update", "origin", "--prune"],
        stdout=subprocess.DEVNULL,
    )
    branches = (
        subprocess.check_output(["git", "branch", "-vv"])
        .decode(sys.stdout.encoding)
        .strip()
        .splitlines()
    )
    for branch in branches:
        if ": gone]" in branch:
            subprocess.run(["git", "branch", "-D", branch.strip().split()[0]])
    subprocess.run(["git", "checkout", "-"])


@cli_git.command()
def init_conf():
    """Initialize local GIT config"""
    return 0


if __name__ == "__main__":
    cli_git.main()
